#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/log/__init__.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 11.06.2020
# Last Modified Date: 11.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

# flake8: noqa: F401
from .AmpelLogger import AmpelLogger, DEBUG, INFO, VERBOSE, SHOUT, WARNING, ERROR
from .LogFlag import LogFlag
